"""AI4 Metadata validator."""

from contextlib import suppress
import importlib.metadata
import pathlib

import enum

__version__ = "2.2.0"


def extract_version() -> str:
    """Return either the version of the package installed."""
    with suppress(FileNotFoundError, StopIteration):
        root_dir = pathlib.Path(__file__).parent.parent.parent
        with open(root_dir / "pyproject.toml", encoding="utf-8") as pyproject_toml:
            version = (
                next(line for line in pyproject_toml if line.startswith("version"))
                .split("=")[1]
                .strip("'\"\n ")
            )
            return f"{version}-dev (at {root_dir})"
    return importlib.metadata.version(__package__ or __name__.split(".", maxsplit=1)[0])


class MetadataVersions(str, enum.Enum):
    """Available versions of the AI4 metadata schema."""

    V1 = "1.0.0"
    V2 = "2.0.0"


_metadata_version_files = {
    MetadataVersions.V1: pathlib.Path(
        pathlib.Path(__file__).parent / "schemata/ai4-apps-v1.0.0.json"
    ),
    MetadataVersions.V2: pathlib.Path(
        pathlib.Path(__file__).parent / "schemata/ai4-apps-v2.0.0.json"
    ),
}

LATEST_METADATA_VERSION = MetadataVersions.V2


def get_latest_version() -> MetadataVersions:
    """Get the latest version of the AI4 metadata schema."""
    return LATEST_METADATA_VERSION


def get_schema(version: MetadataVersions) -> pathlib.Path:
    """Get the schema file path for a given version."""
    return _metadata_version_files[version]



"""Generate an AI4 metadata follwowing schema with empty of with samples."""

import collections
import json
import pathlib
import typing

from ai4_metadata import exceptions


def generate(
    schema: typing.Union[pathlib.Path, str],
    sample_values: bool = False,
    required_only: bool = False,
) -> collections.OrderedDict:
    """Generate an AI4 metadata schema empty of with samples."""
    schema_json = json.load(open(schema, "r"))

    properties = schema_json.get("properties")
    required = schema_json.get("required", [])

    if required_only:
        properties = {k: v for k, v in properties.items() if k in required}

    if not properties:
        raise exceptions.InvalidSchemaError(
            schema, "No definitions found in the schema."
        )

    generated_json: collections.OrderedDict[str, typing.Any] = collections.OrderedDict()

    version = properties.pop("metadata_version").get("example")
    generated_json["metadata_version"] = version

    for key, value in properties.items():
        generated_json[key] = _get_field_value(value, sample_values)

    return generated_json


def _get_field_value(value: dict, sample_values: bool = False) -> typing.Any:
    """Get the value of a field."""
    if value.get("type") == "object":
        required = value.get("required", [])

        properties = value.get("properties", {})
        if required:
            properties = {k: v for k, v in properties.items() if k in required}

        aux = collections.OrderedDict()
        for key, sub_value in properties.items():
            aux[key] = _get_field_value(sub_value, sample_values)
        return aux
    elif value.get("type") == "array":
        if sample_values:
            return value.get("example", [])
        else:
            return []
    else:
        if sample_values:
            return value.get("example", "")
        else:
            return ""



"""Unit tests for the validator module."""

import pathlib
import pytest

from ai4_metadata import exceptions
from ai4_metadata import utils
from ai4_metadata import validate


@pytest.fixture
def valid_schema_file():
    """Fixture for a valid schema file."""
    return pathlib.Path(__file__).parent / "../schemata/ai4-apps-v2.0.0.json"


@pytest.fixture
def invalid_schema_file():
    """Fixture for an invalid schema file."""
    return ""


@pytest.fixture
def valid_instance():
    """Fixture for a valid instance."""
    i = pathlib.Path(__file__).parent / "../../../instances/sample-v2.mods.json"
    return utils.load_json(i)


@pytest.fixture
def valid_instance_file():
    """Fixture for a valid instance."""
    return pathlib.Path(__file__).parent / "../../../instances/sample-v2.mods.json"


@pytest.fixture
def valid_yaml_instance():
    """Fixture for a valid instance."""
    i = pathlib.Path(__file__).parent / "../../../instances/sample-v2.mods.yaml"
    return utils.load_yaml(i)


@pytest.fixture
def valid_yaml_instance_file():
    """Fixture for a valid instance."""
    return pathlib.Path(__file__).parent / "../../../instances/sample-v2.mods.yaml"


@pytest.fixture
def invalid_instance():
    """Fixture for an invalid instance."""
    return {"foo": "bar"}


@pytest.fixture
def invalid_instance_file():
    """Fixture for an invalid instance."""
    return pathlib.Path(__file__).parent / "../../../instances/invalid.json"


@pytest.fixture
def not_found_instance_file():
    """Fixture for an invalid instance."""
    return pathlib.Path(__file__).parent / "foo/bar/baz/instance.json"


def test_validator(valid_schema_file, valid_instance_file):
    """Test the validator with a loaded JSON."""
    assert validate.validate(valid_instance_file, valid_schema_file) is None


def test_validator_file(valid_schema_file, valid_instance):
    """Test the validator with a JSON file."""
    assert validate.validate(valid_instance, valid_schema_file) is None


def test_validator_yaml(valid_schema_file, valid_yaml_instance):
    """Test the validator with a loaded YAML."""
    assert validate.validate(valid_yaml_instance, valid_schema_file) is None


def test_validator_yaml_file(valid_schema_file, valid_yaml_instance_file):
    """Test the validator with a YAML file."""
    assert validate.validate(valid_yaml_instance_file, valid_schema_file) is None


def test_validator_invalid_instance(valid_schema_file, invalid_instance):
    """Test the validator with an invalid instance."""
    with pytest.raises(exceptions.MetadataValidationError):
        validate.validate(invalid_instance, valid_schema_file)


def test_validator_invalid_instance_file(valid_schema_file, invalid_instance_file):
    """Test the validator with an invalid instance file."""
    with pytest.raises(exceptions.MetadataValidationError):
        validate.validate(invalid_instance_file, valid_schema_file)


def test_validator_not_found_instance_file(valid_schema_file, not_found_instance_file):
    """Test the validator with a not found instance file."""
    with pytest.raises(exceptions.FileNotFoundError):
        validate.validate(not_found_instance_file, valid_schema_file)


def test_validator_invalid_schema(invalid_schema_file, valid_instance):
    """Test the validator with an invalid schema."""
    with pytest.raises(exceptions.SchemaValidationError):
        validate.validate(valid_instance, invalid_schema_file)


def test_validator_invallid_all(invalid_schema_file, invalid_instance):
    """Test the validator with an invalid schema and instance."""
    with pytest.raises(exceptions.SchemaValidationError):
        validate.validate(invalid_instance, invalid_schema_file)



"""Main module for AI4 metadata validator."""

import pathlib
from jsonschema import validators
import jsonschema.exceptions
import typing

from ai4_metadata import exceptions
from ai4_metadata import utils


def validate(
    instance: typing.Union[dict, pathlib.Path], schema: typing.Union[dict, pathlib.Path]
) -> None:
    """Validate the schema."""
    if isinstance(instance, pathlib.Path):
        instance_file: typing.Union[str, pathlib.Path] = instance
        try:
            instance = utils.load_json(instance_file)
        except exceptions.InvalidJSONError:
            instance = utils.load_yaml(instance_file)
    else:
        instance_file = "no-file"

    if isinstance(schema, pathlib.Path):
        schema_file: typing.Union[str, pathlib.Path] = schema
        schema = utils.load_json(schema_file)
    else:
        schema_file = "no-file"

    try:
        validator = validators.validator_for(schema)
        validator.check_schema(schema)
    except jsonschema.exceptions.SchemaError as e:
        raise exceptions.SchemaValidationError(schema_file, e)

    try:
        validators.validate(instance, schema)
    except jsonschema.exceptions.ValidationError as e:
        raise exceptions.MetadataValidationError(instance_file, e)



"""Utility functions for the AI4 Metadata utils."""

import json
import os
import pathlib
import typing

import rich
import rich.console
import rich.highlighter
import rich.panel
import yaml

from ai4_metadata import exceptions


def load_json(path: typing.Union[str, pathlib.Path]) -> typing.Dict:
    """Load a JSON from the file f."""
    try:
        file = open(path, "r")
        data = file.read()
        return json.loads(data)
    except FileNotFoundError:
        raise exceptions.FileNotFoundError(path)
    except json.JSONDecodeError as e:
        raise exceptions.InvalidJSONError(path, e)


def load_yaml(path: typing.Union[str, pathlib.Path]) -> typing.Dict:
    """Load a YAML from the file f."""
    try:
        file = open(path, "r")
        data = file.read()
        return yaml.safe_load(data)
    except FileNotFoundError:
        raise exceptions.FileNotFoundError(path)
    except yaml.YAMLError as e:
        raise exceptions.InvalidYAMLError(path, e)


def dump_json(data: typing.Dict, path: typing.Optional[pathlib.Path] = None) -> None:
    """Dump a JSON object to stdout or to path if provided."""
    if path is None:
        print(json.dumps(data, indent=4))
    else:
        with open(path, "w") as f:
            json.dump(data, f, indent=4)


_TERMINAL_WIDTH = os.getenv("TERMINAL_WIDTH")
MAX_WIDTH = int(_TERMINAL_WIDTH) if _TERMINAL_WIDTH else None
ALIGN_ERRORS_PANEL: typing.Literal["left", "center", "right"] = "left"
STYLE_ERRORS_PANEL_BORDER = "bold red"
ALIGN_WARNING_PANEL: typing.Literal["left", "center", "right"] = "left"
STYLE_WARNING_PANEL_BORDER = "bold yellow"
ALIGN_OK_PANEL: typing.Literal["left", "center", "right"] = "left"
STYLE_OK_PANEL_BORDER = "bold green"


def _get_rich_console(stderr: bool = False) -> rich.console.Console:
    return rich.console.Console(
        width=MAX_WIDTH,
        stderr=stderr,
    )


def format_rich_error(error: exceptions.BaseExceptionError) -> None:
    """Format an error using rich."""
    console = _get_rich_console(stderr=True)
    console.print(
        rich.panel.Panel(
            f"{error}",
            title="Error",
            highlight=True,
            border_style=STYLE_ERRORS_PANEL_BORDER,
            title_align=ALIGN_ERRORS_PANEL,
        )
    )


def format_rich_warning(error: typing.Union[str, Exception]) -> None:
    """Format a warning using rich."""
    console = _get_rich_console(stderr=True)
    console.print(
        rich.panel.Panel(
            f"{error}",
            title="Warning",
            highlight=True,
            border_style=STYLE_WARNING_PANEL_BORDER,
            title_align=ALIGN_WARNING_PANEL,
        )
    )


def format_rich_ok(message: str) -> None:
    """Format a message using rich."""
    console = _get_rich_console(stderr=False)
    console.print(
        rich.panel.Panel(
            f"{message}",
            title="Success",
            highlight=True,
            border_style=STYLE_OK_PANEL_BORDER,
            title_align=ALIGN_OK_PANEL,
        )
    )



"""Custom exceptions for the metadata schema validation tool."""

import pathlib
import typing

import jsonschema.exceptions


class BaseExceptionError(Exception):
    """Base exception for the metadata schema validation tool."""

    message = "An error occurred."


class FileNotFoundError(BaseExceptionError):
    """Exception raised when a file is not found."""

    message = "File '{f}' not found."

    def __init__(self, f: typing.Union[str, pathlib.Path]):
        """Initialize the exception."""
        self.f = f
        super().__init__(self.message.format(f=f))


class InvalidFileError(BaseExceptionError):
    """Exception raised when a file is invalid."""

    message = "Error loading file '{f}': {e}"

    def __init__(self, f: typing.Union[str, pathlib.Path], e: Exception):
        """Initialize the exception."""
        self.f = f
        self.e = e
        super().__init__(self.message.format(f=f, e=e))


class InvalidJSONError(InvalidFileError):
    """Exception raised when a JSON file is invalid."""

    message = "Error loading JSON file '{f}': {e}"


class InvalidYAMLError(InvalidFileError):
    """Exception raised when a JSON/YAML file is invalid."""

    message = "Error loading YAML file '{f}': {e}"


class InvalidSchemaError(BaseExceptionError):
    """Exception raised when a schema is invalid."""

    message = "Schema file '{schema_file}' is invalid: {error}"

    def __init__(self, schema_file: typing.Union[str, pathlib.Path], error: str):
        """Initialize the exception."""
        super().__init__(self.message.format(schema_file=schema_file, error=error))


class SchemaValidationError(BaseExceptionError):
    """Exception raised when a schema is invalid."""

    message = "Error validating schema '{schema_file}': {e}"

    def __init__(self, schema_file: typing.Union[str, pathlib.Path], e: Exception):
        """Initialize the exception."""
        self.e = e
        super().__init__(self.message.format(schema_file=schema_file, e=e))


class MetadataValidationError(BaseExceptionError):
    """Exception raised when a metadata file is invalid."""

    message = "Error validating instance '{instance_file}': {e}"

    def __init__(
        self,
        instance_file: typing.Union[str, pathlib.Path],
        e: jsonschema.exceptions.ValidationError,
    ):
        """Initialize the exception."""
        self.instance_file = instance_file
        self.e = e
        super().__init__(self.message.format(instance_file=instance_file, e=e.message))



"""Command line interface for the ai4-metadata package."""

import pathlib
from typing_extensions import Annotated, Optional
from typing import List
import warnings

import typer

import ai4_metadata
from ai4_metadata import generate
from ai4_metadata import exceptions
from ai4_metadata import migrate
from ai4_metadata import utils
from ai4_metadata import validate

app = typer.Typer()


@app.command(name="migrate")
def _migrate(
    v1_metadata: Annotated[
        pathlib.Path, typer.Argument(help="AI4 application metadata file to migrate.")
    ],
    output: Annotated[
        Optional[pathlib.Path],
        typer.Option("--output", "-o", help="Output file for migrated metadata."),
    ] = None,
):
    """Migrate an AI4 metadata file from V1 to the latest V2 version."""
    v1_schema = ai4_metadata.get_schema(ai4_metadata.MetadataVersions.V1)
    validate.validate(v1_metadata, v1_schema)

    # Migrate metadata
    v2_metadata = migrate.migrate(v1_metadata)
    v2_version = ai4_metadata.get_latest_version()
    v2_schema = ai4_metadata.get_schema(v2_version)
    validate.validate(v2_metadata, v2_schema)

    # Write out the migrated metadata
    utils.dump_json(v2_metadata, output)
    if output:
        utils.format_rich_ok(
            f"V1 metadata '{v1_metadata}' migrated to version {v2_version} and "
            f"stored in '{output}'",
        )


@app.command(name="validate")
def _validate(
    instances: Annotated[
        List[pathlib.Path],
        typer.Argument(
            help="AI4 application metadata file to validate. The file can "
            "be repeated to validate multiple files. Supported formats are "
            "JSON and YAML."
        ),
    ],
    schema: Annotated[
        Optional[pathlib.Path],
        typer.Option(help="AI4 application metadata schema file to use."),
    ] = None,
    metadata_version: Annotated[
        ai4_metadata.MetadataVersions,
        typer.Option(help="AI4 application metadata version."),
    ] = ai4_metadata.get_latest_version(),
    quiet: Annotated[
        bool, typer.Option("--quiet", "-q", help="Suppress output for valid instances.")
    ] = False,
):
    """Validate an AI4 metadata file against the AI4 metadata schema.

    This command receives an AI4 metadata file and validates it against a
    given version of the metadata schema. By default it will check against the latest
    metadata version.

    If the metadata is not valid it will exit with .

    If you provide the --shema option, it will override the --metadata-version option.
    """
    schema_file = schema or ai4_metadata.get_schema(metadata_version)

    exit_code = 0
    for instance_file in instances:
        try:
            validate.validate(instance_file, schema_file)
        except exceptions.FileNotFoundError as e:
            utils.format_rich_error(e)
            typer.Exit(2)
        except exceptions.InvalidFileError as e:
            utils.format_rich_error(e)
            typer.Exit(2)
        except exceptions.SchemaValidationError as e:
            utils.format_rich_error(e)
            typer.Exit(3)
        except exceptions.MetadataValidationError as e:
            utils.format_rich_error(e)
            exit_code = 1
        else:
            if not quiet:
                utils.format_rich_ok(
                    f"'{instance_file}' is valid for version {metadata_version.value}"
                )

    raise typer.Exit(code=exit_code)


@app.command(name="generate")
def _generate(
    metadata_version: Annotated[
        ai4_metadata.MetadataVersions,
        typer.Option(help="AI4 application metadata version."),
    ] = ai4_metadata.get_latest_version(),
    sample_values: Annotated[
        bool, typer.Option("--sample-values", help="Generate sample values.")
    ] = False,
    required: Annotated[
        bool, typer.Option("--required-only", help="Include only required fields.")
    ] = False,
    output: Annotated[
        Optional[pathlib.Path],
        typer.Option("--output", "-o", help="Output file for generated metadata."),
    ] = None,
):
    """Generate an AI4 metadata schema."""
    schema = ai4_metadata.get_schema(metadata_version)

    try:
        generated_json = generate.generate(schema, sample_values, required)
    except exceptions.InvalidSchemaError as e:
        utils.format_rich_error(e)
        raise typer.Exit(1)

    utils.dump_json(generated_json, output)

    validate.validate(generated_json, schema)

    if output:
        utils.format_rich_ok(
            f"Sample file stored in '{output}' for version {metadata_version.value}"
        )


def version_callback(value: bool):
    """Return the version for the --version option."""
    if value:
        typer.echo(ai4_metadata.extract_version())
        raise typer.Exit()


@app.callback()
def version(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        is_eager=True,
        help="Print the version and exit",
    )
):
    """Show version and exit."""
    pass


def migrate_main():
    """Run the migration command as an independent script."""
    # NOTE(aloga): This is a workaround to be able to provide the command as a separate
    # script, in order to be compatible with previous versions of the package. However,
    # this will be not be supported in the next major version of the package, therfore
    # we mark it as deprecated and raise a warining
    msg = (
        "The 'ai4-metadata-migrate' command is deprecated and will be removed "
        "in the next major version of the package, please use 'ai4-metadata migrate' "
        "instead."
    )
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    utils.format_rich_warning(DeprecationWarning(msg))
    typer.run(_migrate)


def validate_main():
    """Run the validation command as an independent script."""
    # NOTE(aloga): This is a workaround to be able to provide the command as a separate
    # script, in order to be compatible with previous versions of the package. However,
    # this will be not be supported in the next major version of the package, therfore
    # we mark it as deprecated and raise a warining
    msg = (
        "The 'ai4-metadata-validator' command is deprecated and will be removed "
        "in the next major version of the package, please use 'ai4-metadata validate' "
        "instead."
    )
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    utils.format_rich_warning(DeprecationWarning(msg))
    typer.run(_validate)



"""Migrate metadata from V1 to V2."""

import collections
import datetime
import pathlib
import typing

import ai4_metadata
from ai4_metadata import utils


def migrate(instance_file: pathlib.Path) -> collections.OrderedDict:
    """Try to migrate metadata from V1 to latest V2."""
    v1_metadata = utils.load_json(instance_file)

    v2: collections.OrderedDict[str, typing.Any] = collections.OrderedDict()

    v2["metadata_version"] = ai4_metadata.get_latest_version().value
    v2["title"] = v1_metadata.get("title")
    v2["summary"] = v1_metadata.get("summary")
    v2["description"] = " ".join(v1_metadata.get("description", []))
    v2["dates"] = {
        "created": v1_metadata.get("date_creation"),
        "updated": datetime.datetime.now().strftime("%Y-%m-%d"),
    }
    v2["links"] = {
        "source_code": v1_metadata.get("sources", {}).get("code"),
        "docker_image": v1_metadata.get("sources", {}).get("docker_registry_repo"),
    }
    v2["tags"] = v1_metadata.get("keywords", [])
    v2["tasks"] = []
    v2["categories"] = []
    v2["libraries"] = []

    # Now move things, if present, into links
    if v1_metadata.get("doi"):
        v2["links"]["doi"] = v1_metadata.get("doi")
    if v1_metadata.get("sources", {}).get("zenodo_doi"):
        v2["links"]["zenodo_doi"] = v1_metadata.get("sources", {}).get("zenodo_doi")
    if v1_metadata.get("sources", {}).get("pre_trained_weights"):
        v2["links"]["weights"] = v1_metadata.get("sources", {}).get(
            "pre_trained_weights"
        )
    if v1_metadata.get("sources", {}).get("ai4_template"):
        v2["links"]["ai4_template"] = v1_metadata.get("sources", {}).get("ai4_template")

    if v1_metadata.get("dataset_url"):
        v2["links"]["dataset_url"] = v1_metadata.get("dataset_url")
    if v1_metadata.get("training_files_url"):
        v2["links"]["training_files_url"] = v1_metadata.get("training_files_url")
    if v1_metadata.get("cite_url"):
        v2["links"]["cite_url"] = v1_metadata.get("cite_url")

    # Try to infer some some more information on libraries and categories
    kw = [k.lower() for k in v1_metadata.get("keywords", [])]
    if "tensorflow" in kw:
        v2["libraries"].append("TensorFlow")
    if "pytorch" in kw:
        v2["libraries"].append("PyTorch")
    if "keras" in kw:
        v2["libraries"].append("Keras")
    if "scikit-learn" in kw:
        v2["libraries"].append("Scikit-Learn")

    if "trainable" in kw:
        v2["categories"].append("AI4 trainable")
    if "inference" in kw:
        v2["categories"].append("AI4 inference")
        v2["categories"].append("AI4 pre trained")

    return v2
